"""Debug utilities for WhispyWyser."""
from __future__ import annotations

import json
import logging
import os
from dataclasses import asdict, is_dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from .homeassistant.models import Entity, Things

_LOGGER = logging.getLogger(__name__)

class EntityDebugLogger:
    """Helper class to log entity information for debugging."""
    
    def __init__(self, output_dir: Union[str, Path] = "./debug"):
        """Initialize the debug logger.
        
        Args:
            output_dir: Directory to store debug files
        """
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.entities_file = self.output_dir / "entities.json"
        _LOGGER.info("Entity debug logging enabled. Output directory: %s", self.output_dir)
    
    def log_entities(self, things: Things) -> None:
        """Log discovered entities to a JSON file.
        
        Args:
            things: The Things object containing entities and areas
        """
        try:
            # Convert dataclasses to dictionaries
            entities_data = [self._entity_to_dict(e) for e in things.entities]
            
            # Prepare the output data
            output = {
                "timestamp": datetime.utcnow().isoformat(),
                "entities": entities_data,
                "areas": [{"names": a.names, "area_id": a.area_id} for a in things.areas]
            }
            
            # Write to file
            with open(self.entities_file, "w", encoding="utf-8") as f:
                json.dump(output, f, indent=2, ensure_ascii=False)
                
            _LOGGER.debug("Logged %d entities to %s", len(entities_data), self.entities_file)
            
            self._generate_spell_training(things)
            
        except Exception as e:
            _LOGGER.error("Failed to log entities: %s", e, exc_info=True)
    
    def _generate_spell_training(self, things: Things) -> None:
        """Generate spell training data from entities and intents."""
        try:
            # Look for intents in standard locations
            intent_dirs = [
                Path("intents"),  # Local intents
                Path("/config/intents"),  # Home Assistant addon
                Path(__file__).parent.parent / "intents"  # Package intents
            ]
            
            intent_dir = None
            for dir_path in intent_dirs:
                if dir_path.exists() and any(dir_path.glob("*.yaml")):
                    intent_dir = dir_path
                    _LOGGER.debug("Found intents in %s", intent_dir)
                    break
            
            training_file = generate_spell_training(
                things=things,
                output_dir=self.output_dir / "spell_training",
                intent_dir=intent_dir,
                language="en"  # TODO: Make configurable
            )
            _LOGGER.info("Generated spell training data in %s", training_file)
            
        except Exception as e:
            _LOGGER.error("Failed to generate spell training data: %s", e, exc_info=True)
    
    def _entity_to_dict(self, entity: Entity) -> Dict[str, Any]:
        """Convert an Entity to a dictionary, handling dataclass fields."""
        result = {
            "entity_id": getattr(entity, "entity_id", "unknown"),
            "names": entity.names,
            "domain": entity.domain,
        }
        
        # Add all fields that start with 'supports_'
        for field in dir(entity):
            if field.startswith("supports_") or field.startswith("light_") or field.startswith("fan_") or field.startswith("cover_") or field.startswith("media_player_"):
                value = getattr(entity, field, None)
                if value is not None:
                    result[field] = value
        
        return result

def setup_debug_logging(debug: bool = False, output_dir: Optional[str] = None) -> Optional[EntityDebugLogger]:
    """Set up debug logging if enabled.
    
    Args:
        debug: Whether to enable debug logging
        output_dir: Directory to store debug files (default: ./debug)
        
    Returns:
        Configured EntityDebugLogger instance if debug is True, else None
    """
    if not debug:
        return None
        
    # Set up logging level
    logging.basicConfig(level=logging.DEBUG)
    
    # Create and return debug logger
    return EntityDebugLogger(output_dir=output_dir or "./debug")



"""Event handler for clients of the server."""
import argparse
import asyncio
import logging
import os
import tempfile
import wave
from typing import Optional

import faster_whisper
from wyoming.asr import Transcribe, Transcript
from wyoming.audio import AudioChunk, AudioStop
from wyoming.event import Event
from wyoming.info import Describe, Info
from wyoming.server import AsyncEventHandler

_LOGGER = logging.getLogger(__name__)


class FasterWhisperEventHandler(AsyncEventHandler):
    """Event handler for clients."""

    def __init__(
        self,
        wyoming_info: Info,
        cli_args: argparse.Namespace,
        model: faster_whisper.WhisperModel,
        model_lock: asyncio.Lock,
        *args,
        initial_prompt: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.cli_args = cli_args
        self.wyoming_info_event = wyoming_info.event()
        self.model = model
        self.model_lock = model_lock
        self.initial_prompt = initial_prompt
        self._language = self.cli_args.language
        self._wav_dir = tempfile.TemporaryDirectory()
        self._wav_path = os.path.join(self._wav_dir.name, "speech.wav")
        self._wav_file: Optional[wave.Wave_write] = None

    async def handle_event(self, event: Event) -> bool:
        if AudioChunk.is_type(event.type):
            chunk = AudioChunk.from_event(event)

            if self._wav_file is None:
                self._wav_file = wave.open(self._wav_path, "wb")
                self._wav_file.setframerate(chunk.rate)
                self._wav_file.setsampwidth(chunk.width)
                self._wav_file.setnchannels(chunk.channels)

            self._wav_file.writeframes(chunk.audio)
            return True

        if AudioStop.is_type(event.type):
            _LOGGER.debug(
                "Audio stopped. Transcribing with initial prompt=%s",
                self.initial_prompt,
            )
            assert self._wav_file is not None

            self._wav_file.close()
            self._wav_file = None

            async with self.model_lock:
                segments, _info = self.model.transcribe(
                    self._wav_path,
                    beam_size=self.cli_args.beam_size,
                    language=self._language,
                    initial_prompt=self.initial_prompt,
                )

            text = " ".join(segment.text for segment in segments)
            _LOGGER.info(text)

            await self.write_event(Transcript(text=text).event())
            _LOGGER.debug("Completed request")

            # Reset
            self._wav_path = bytes()
            self._language = self.cli_args.language

            return False

        if Transcribe.is_type(event.type):
            transcribe = Transcribe.from_event(event)
            if transcribe.language:
                self._language = transcribe.language
                _LOGGER.debug("Language set to %s", transcribe.language)
            return True

        if Describe.is_type(event.type):
            await self.write_event(self.wyoming_info_event)
            _LOGGER.debug("Sent info")
            return True

        return True



"""Spell checking functionality for WhispyWyser."""
from .checker import JamSpellChecker, SpellCheckResponse
from .trainer import SpellTrainer, generate_spell_training

__all__ = [
    'JamSpellChecker',
    'SpellCheckResponse',
    'SpellTrainer',
    'generate_spell_training',
]



"""Code for Whisper transcription using HuggingFace's transformers library."""
import asyncio
import logging
import os
import tempfile
import wave
from pathlib import Path
from typing import Optional, Union

import torch
from transformers import AutoModelForSpeechSeq2Seq, AutoProcessor
from wyoming.asr import Transcribe, Transcript
from wyoming.audio import AudioChunk, AudioStop
from wyoming.event import Event
from wyoming.info import Describe, Info
from wyoming.server import AsyncEventHandler

_LOGGER = logging.getLogger(__name__)


class TransformersWhisperModel:
    """Wrapper for HuggingFace transformers Whisper model."""

    def __init__(
        self,
        model_id: str,
        cache_dir: Optional[Union[str, Path]] = None,
        local_files_only: bool = False,
    ) -> None:
        """Initialize Whisper model."""
        self.processor = AutoProcessor.from_pretrained(
            model_id, cache_dir=cache_dir, local_files_only=local_files_only
        )
        self.model = AutoModelForSpeechSeq2Seq.from_pretrained(
            model_id, cache_dir=cache_dir, local_files_only=local_files_only
        )
        self.model.eval()

    def transcribe(
        self,
        wav_path: Union[str, Path],
        beam_size: int = 5,
        language: Optional[str] = None,
    ) -> str:
        """Returns transcription for WAV file.

        WAV file must be 16Khz 16-bit mono audio.
        """
        wav_file: wave.Wave_read = wave.open(str(wav_path), "rb")
        with wav_file:
            assert wav_file.getframerate() == 16000, "Sample rate must be 16Khz"
            assert wav_file.getsampwidth() == 2, "Width must be 16-bit (2 bytes)"
            assert wav_file.getnchannels() == 1, "Audio muts be mono"
            audio_bytes = wav_file.readframes(wav_file.getnframes())

        audio_tensor = (
            torch.frombuffer(audio_bytes, dtype=torch.int16).float() / 32768.0
        )

        inputs = self.processor(audio_tensor, sampling_rate=16000, return_tensors="pt")
        generate_args = {**inputs, "num_beams": beam_size}

        if language:
            generate_args["forced_decoder_ids"] = self.processor.get_decoder_prompt_ids(
                language=language, task="transcribe"
            )

        with torch.no_grad():
            # Ignore warning about attention_mask because we're only doing a single utterance.
            generated_ids = self.model.generate(**generate_args)
            transcription = self.processor.batch_decode(
                generated_ids, skip_special_tokens=True
            )[0]

        return transcription


class TransformersWhisperEventHandler(AsyncEventHandler):
    """Event handler for clients."""

    def __init__(
        self,
        wyoming_info: Info,
        language: Optional[str],
        beam_size: int,
        model: TransformersWhisperModel,
        model_lock: asyncio.Lock,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.wyoming_info_event = wyoming_info.event()
        self.model = model
        self.model_lock = model_lock
        self._beam_size = beam_size
        self._language = language
        self._wav_dir = tempfile.TemporaryDirectory()
        self._wav_path = os.path.join(self._wav_dir.name, "speech.wav")
        self._wav_file: Optional[wave.Wave_write] = None

    async def handle_event(self, event: Event) -> bool:
        if AudioChunk.is_type(event.type):
            chunk = AudioChunk.from_event(event)

            if self._wav_file is None:
                self._wav_file = wave.open(self._wav_path, "wb")
                self._wav_file.setframerate(chunk.rate)
                self._wav_file.setsampwidth(chunk.width)
                self._wav_file.setnchannels(chunk.channels)

            self._wav_file.writeframes(chunk.audio)
            return True

        if AudioStop.is_type(event.type):
            _LOGGER.debug(
                "Audio stopped. Transcribing with language=%s", self._language
            )
            assert self._wav_file is not None

            self._wav_file.close()
            self._wav_file = None

            async with self.model_lock:
                text = self.model.transcribe(
                    self._wav_path,
                    beam_size=self._beam_size,
                    language=self._language,
                )

            _LOGGER.info(text)

            await self.write_event(Transcript(text=text).event())
            _LOGGER.debug("Completed request")

            # Reset
            self._wav_file = bytes()
            self._language = self._language

            return False

        if Transcribe.is_type(event.type):
            transcribe = Transcribe.from_event(event)
            if transcribe.language:
                self._language = transcribe.language
                _LOGGER.debug("Language set to %s", transcribe.language)
            return True

        if Describe.is_type(event.type):
            await self.write_event(self.wyoming_info_event)
            _LOGGER.debug("Sent info")
            return True

        return True



"""Generate training data for JamSpell spell checker."""
from __future__ import annotations

import json
import logging
import re
import subprocess
import sys
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Set, Tuple, Union

from .homeassistant.models import Entity, Things

_LOGGER = logging.getLogger(__name__)

class SpellTrainer:
    """Generates training data for JamSpell spell checker using Hassil intents."""
    
    def __init__(self, output_dir: Union[str, Path] = "./spell_train"):
        """Initialize the spell trainer."""
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.sentences: Set[str] = set()
        _LOGGER.info("Spell trainer initialized. Output directory: %s", self.output_dir)
    
    def generate_from_entities(self, things: Things) -> None:
        """Generate training sentences from Home Assistant entities."""
        # Add entity names as training data
        for entity in things.entities:
            for name in entity.names:
                self.sentences.add(name.lower())
        
        # Add area names
        for area in things.areas:
            for name in area.names:
                self.sentences.add(name.lower())
    
    def generate_from_hassil(self, intent_dir: Union[str, Path], language: str = "en") -> None:
        """Generate training sentences using Hassil's sample sentences.
        
        Args:
            intent_dir: Directory containing Hassil intent YAML files
            language: Language code (default: en)
        """
        try:
            from hassil.sample import sample_combinations
            
            intent_dir = Path(intent_dir)
            if not intent_dir.exists():
                _LOGGER.warning("Intent directory not found: %s", intent_dir)
                return
                
            # Sample sentences from Hassil intents
            for intent_file in intent_dir.glob("*.yaml"):
                try:
                    for _ in range(10):  # Generate 10 variations per intent
                        result = subprocess.run(
                            [sys.executable, "-m", "hassil.sample", str(intent_file), "--language", language],
                            capture_output=True,
                            text=True,
                            check=True
                        )
                        
                        # Parse the output
                        for line in result.stdout.splitlines():
                            try:
                                data = json.loads(line)
                                if "text" in data:
                                    self.sentences.add(data["text"].lower())
                            except json.JSONDecodeError:
                                self.sentences.add(line.strip().lower())
                                
                except subprocess.CalledProcessError as e:
                    _LOGGER.error("Error sampling from %s: %s", intent_file, e.stderr)
                    
        except ImportError:
            _LOGGER.warning("Hassil not installed, skipping intent-based training data")
    
    def save_training_data(self, filename: str = "training.txt") -> Path:
        """Save generated sentences to a file.
        
        Args:
            filename: Output filename
            
        Returns:
            Path to the generated training file
        """
        output_file = self.output_dir / filename
        
        with open(output_file, "w", encoding="utf-8") as f:
            for sentence in sorted(self.sentences):
                if sentence.strip():  # Skip empty lines
                    f.write(f"{sentence}\n")
        
        _LOGGER.info("Saved %d training sentences to %s", len(self.sentences), output_file)
        return output_file

def generate_spell_training(
    things: Things, 
    output_dir: Union[str, Path],
    intent_dir: Optional[Union[str, Path]] = None,
    language: str = "en"
) -> Path:
    """Generate spell training data from Home Assistant entities and Hassil intents.
    
    Args:
        things: The Things object containing entities and areas
        output_dir: Directory to save training data
        intent_dir: Optional directory containing Hassil intent YAML files
        language: Language code (default: en)
        
    Returns:
        Path to the generated training file
    """
    trainer = SpellTrainer(output_dir=output_dir)
    
    # Generate from entity names and areas
    trainer.generate_from_entities(things)
    
    # Generate from Hassil intents if available
    if intent_dir:
        trainer.generate_from_hassil(intent_dir, language=language)
    
    # Save the training data
    return trainer.save_training_data()



"""Data models for Home Assistant integration."""
from __future__ import annotations

import hashlib
from dataclasses import dataclass, field, fields
from typing import Any, Dict, List, Optional

@dataclass
class Entity:
    """Home Assistant entity."""

    entity_id: str
    names: List[str]
    domain: str
    
    # Domain-specific features
    light_supports_color: Optional[bool] = None
    light_supports_brightness: Optional[bool] = None
    fan_supports_speed: Optional[bool] = None
    cover_supports_position: Optional[bool] = None
    media_player_supports_pause: Optional[bool] = None
    media_player_supports_volume_set: Optional[bool] = None
    media_player_supports_next_track: Optional[bool] = None
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for this entity."""
        if not self._hash:
            hasher = hashlib.sha256()
            hasher.update(self.entity_id.encode("utf-8"))
            hasher.update(self.domain.encode("utf-8"))

            for supports_field in fields(self):
                if "supports" not in supports_field.name:
                    continue

                supports_value = getattr(self, supports_field.name)
                if supports_value is None:
                    continue

                hasher.update(f"{supports_field.name}={supports_value}".encode("utf-8"))

            for name in sorted(self.names):
                hasher.update(name.encode("utf-8"))

            self._hash = hasher.hexdigest()

        return self._hash

@dataclass
class Area:
    """Home Assistant area."""

    area_id: str
    names: List[str]
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for this area."""
        if not self._hash:
            hasher = hashlib.sha256()
            hasher.update(self.area_id.encode("utf-8"))

            for name in sorted(self.names):
                hasher.update(name.encode("utf-8"))

            self._hash = hasher.hexdigest()

        return self._hash

@dataclass
class Things:
    """Exposed things in Home Assistant."""

    entities: List[Entity] = field(default_factory=list)
    areas: List[Area] = field(default_factory=list)
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for all the things."""
        if not self._hash:
            hasher = hashlib.sha256()
            
            for entity in sorted(self.entities, key=lambda e: e.entity_id):
                hasher.update(entity.get_hash().encode("utf-8"))
                
            for area in sorted(self.areas, key=lambda a: a.area_id):
                hasher.update(area.get_hash().encode("utf-8"))
                
            self._hash = hasher.hexdigest()
            
        return self._hash

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dictionary for JSON serialization."""
        return {
            "entities": [
                {
                    "entity_id": e.entity_id,
                    "names": e.names,
                    "domain": e.domain,
                    "light_supports_color": e.light_supports_color,
                    "light_supports_brightness": e.light_supports_brightness,
                    "fan_supports_speed": e.fan_supports_speed,
                    "cover_supports_position": e.cover_supports_position,
                    "media_player_supports_pause": e.media_player_supports_pause,
                    "media_player_supports_volume_set": e.media_player_supports_volume_set,
                    "media_player_supports_next_track": e.media_player_supports_next_track,
                }
                for e in self.entities
            ],
            "areas": [
                {"area_id": a.area_id, "names": a.names} for a in self.areas
            ],
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'Things':
        """Create from a dictionary."""
        things = cls()
        
        # Add entities
        for entity_data in data.get("entities", []):
            entity = Entity(
                entity_id=entity_data["entity_id"],
                names=entity_data["names"],
                domain=entity_data["domain"],
                light_supports_color=entity_data.get("light_supports_color"),
                light_supports_brightness=entity_data.get("light_supports_brightness"),
                fan_supports_speed=entity_data.get("fan_supports_speed"),
                cover_supports_position=entity_data.get("cover_supports_position"),
                media_player_supports_pause=entity_data.get("media_player_supports_pause"),
                media_player_supports_volume_set=entity_data.get("media_player_supports_volume_set"),
                media_player_supports_next_track=entity_data.get("media_player_supports_next_track"),
            )
            things.entities.append(entity)
            
        # Add areas
        for area_data in data.get("areas", []):
            area = Area(
                area_id=area_data["area_id"],
                names=area_data["names"]
            )
            things.areas.append(area)
            
        return things



"""Home Assistant API client for WhispyWyser."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import aiohttp

from .client import HomeAssistantClient, HomeAssistantError
from .models import Area, Entity, Things

__all__ = [
    "HomeAssistantClient",
    "HomeAssistantError",
    "Area",
    "Entity",
    "Things",
    "get_hass_info",
]

async def get_hass_info(
    url: str,
    token: str,
    session: Optional[aiohttp.ClientSession] = None,
) -> Things:
    """Get exposed entities and areas from Home Assistant.
    
    Args:
        url: Base URL of the Home Assistant instance
        token: Long-lived access token
        session: Optional aiohttp ClientSession
        
    Returns:
        Things object containing entities and areas
        
    Raises:
        HomeAssistantError: If there's an error communicating with Home Assistant
    """
    from .client import get_hass_info as _get_hass_info
    return await _get_hass_info(url, token, session)



"""Home Assistant integration for WhispyWyser."""
from __future__ import annotations

import asyncio
import hashlib
import logging
import re
from collections.abc import AsyncGenerator
from dataclasses import dataclass, field, fields
from typing import Any, Dict, List, Optional, Set, Union

import aiohttp

_LOGGER = logging.getLogger(__name__)

RGB_MODES = {"hs", "rgb", "rgbw", "rgbww", "xy"}
BRIGHTNESS_MODES = RGB_MODES | {"brightness", "white"}
FAN_SET_SPEED = 1
COVER_SET_POSITION = 4
MEDIA_PLAYER_PAUSE = 1
MEDIA_PLAYER_VOLUME_SET = 4
MEDIA_PLAYER_NEXT_TRACK = 32

class HomeAssistantError(Exception):
    """Base exception for Home Assistant errors."""
    pass

@dataclass
class Entity:
    """Home Assistant entity."""

    entity_id: str
    names: List[str]
    domain: str
    
    # Domain-specific features
    light_supports_color: Optional[bool] = None
    light_supports_brightness: Optional[bool] = None
    fan_supports_speed: Optional[bool] = None
    cover_supports_position: Optional[bool] = None
    media_player_supports_pause: Optional[bool] = None
    media_player_supports_volume_set: Optional[bool] = None
    media_player_supports_next_track: Optional[bool] = None
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for this entity."""
        if not self._hash:
            hasher = hashlib.sha256()
            hasher.update(self.entity_id.encode("utf-8"))
            hasher.update(self.domain.encode("utf-8"))

            for supports_field in fields(self):
                if "supports" not in supports_field.name:
                    continue

                supports_value = getattr(self, supports_field.name)
                if supports_value is None:
                    continue

                hasher.update(f"{supports_field.name}={supports_value}".encode("utf-8"))

            for name in sorted(self.names):
                hasher.update(name.encode("utf-8"))

            self._hash = hasher.hexdigest()

        return self._hash

@dataclass
class Area:
    """Home Assistant area."""

    area_id: str
    names: List[str]
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for this area."""
        if not self._hash:
            hasher = hashlib.sha256()
            hasher.update(self.area_id.encode("utf-8"))

            for name in sorted(self.names):
                hasher.update(name.encode("utf-8"))

            self._hash = hasher.hexdigest()

        return self._hash

@dataclass
class Things:
    """Exposed things in Home Assistant."""

    entities: List[Entity] = field(default_factory=list)
    areas: List[Area] = field(default_factory=list)
    _hash: str = ""

    def get_hash(self) -> str:
        """Get a stable hash for all the things."""
        if not self._hash:
            hasher = hashlib.sha256()
            
            for entity in sorted(self.entities, key=lambda e: e.entity_id):
                hasher.update(entity.get_hash().encode("utf-8"))
                
            for area in sorted(self.areas, key=lambda a: a.area_id):
                hasher.update(area.get_hash().encode("utf-8"))
                
            self._hash = hasher.hexdigest()
            
        return self._hash

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dictionary for JSON serialization."""
        return {
            "entities": [
                {
                    "entity_id": e.entity_id,
                    "names": e.names,
                    "domain": e.domain,
                    "light_supports_color": e.light_supports_color,
                    "light_supports_brightness": e.light_supports_brightness,
                    "fan_supports_speed": e.fan_supports_speed,
                    "cover_supports_position": e.cover_supports_position,
                    "media_player_supports_pause": e.media_player_supports_pause,
                    "media_player_supports_volume_set": e.media_player_supports_volume_set,
                    "media_player_supports_next_track": e.media_player_supports_next_track,
                }
                for e in self.entities
            ],
            "areas": [
                {"area_id": a.area_id, "names": a.names} for a in self.areas
            ],
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'Things':
        """Create from a dictionary."""
        things = cls()
        
        # Add entities
        for entity_data in data.get("entities", []):
            entity = Entity(
                entity_id=entity_data["entity_id"],
                names=entity_data["names"],
                domain=entity_data["domain"],
                light_supports_color=entity_data.get("light_supports_color"),
                light_supports_brightness=entity_data.get("light_supports_brightness"),
                fan_supports_speed=entity_data.get("fan_supports_speed"),
                cover_supports_position=entity_data.get("cover_supports_position"),
                media_player_supports_pause=entity_data.get("media_player_supports_pause"),
                media_player_supports_volume_set=entity_data.get("media_player_supports_volume_set"),
                media_player_supports_next_track=entity_data.get("media_player_supports_next_track"),
            )
            things.entities.append(entity)
            
        # Add areas
        for area_data in data.get("areas", []):
            area = Area(
                area_id=area_data["area_id"],
                names=area_data["names"]
            )
            things.areas.append(area)
            
        return things

class HomeAssistantClient:
    """Client for interacting with Home Assistant's API."""
    
    def __init__(
        self,
        url: str,
        token: str,
        session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        """Initialize the Home Assistant client.
        
        Args:
            url: Base URL of the Home Assistant instance (e.g., 'http://homeassistant:8123')
            token: Long-lived access token
            session: Optional aiohttp ClientSession
        """
        self._base_url = url.rstrip('/')
        self._token = token
        self._headers = {
            'Authorization': f'Bearer {token}',
            'Content-Type': 'application/json',
        }
        self._session = session or aiohttp.ClientSession()
        self._websocket = None
        self._message_id = 1
        self._listeners = {}
        self._event_listeners = {}
        
    async def __aenter__(self):
        await self.connect()
        return self
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        
    async def connect(self) -> None:
        """Connect to Home Assistant API and set up WebSocket."""
        if self._session.closed:
            self._session = aiohttp.ClientSession()
            
        await self._setup_websocket()
            
    async def close(self) -> None:
        """Close the connection and clean up resources."""
        if self._websocket is not None:
            await self._websocket.close()
            self._websocket = None
            
        if not self._session.closed:
            await self._session.close()
            
    async def _setup_websocket(self) -> None:
        """Set up WebSocket connection to Home Assistant."""
        ws_url = self._base_url.replace('http', 'ws') + '/api/websocket'
        
        try:
            self._websocket = await self._session.ws_connect(ws_url, heartbeat=55)
            auth_required = await self._websocket.receive_json()
            
            if auth_required['type'] != 'auth_required':
                raise HomeAssistantError('Unexpected message during WebSocket auth')
                
            await self._websocket.send_json({
                'type': 'auth',
                'access_token': self._token
            })
            
            auth_result = await self._websocket.receive_json()
            if auth_result['type'] != 'auth_ok':
                raise HomeAssistantError(f'Authentication failed: {auth_result}')
                
            # Start listening for events
            asyncio.create_task(self._websocket_listener())
            
        except Exception as err:
            _LOGGER.error('WebSocket connection failed: %s', err)
            raise HomeAssistantError(f'WebSocket connection failed: {err}') from err
            
    async def _websocket_listener(self) -> None:
        """Listen for WebSocket messages and dispatch to handlers."""
        try:
            async for msg in self._websocket:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    data = msg.json()
                    await self._handle_websocket_message(data)
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    _LOGGER.warning('WebSocket connection closed')
                    break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    _LOGGER.error('WebSocket error: %s', self._websocket.exception())
                    break
                    
        except Exception as err:
            _LOGGER.error('WebSocket listener error: %s', err)
        finally:
            self._websocket = None
            
    async def _handle_websocket_message(self, data: Dict[str, Any]) -> None:
        """Handle incoming WebSocket messages."""
        if 'id' in data and data['id'] in self._listeners:
            future = self._listeners.pop(data['id'])
            if not future.done():
                if data.get('success', True):
                    future.set_result(data.get('result'))
                else:
                    future.set_exception(HomeAssistantError(data.get('error', {}).get('message', 'Unknown error')))
                    
        elif data.get('type') == 'event':
            event_type = data.get('event', {}).get('event_type')
            if event_type in self._event_listeners:
                for callback in self._event_listeners[event_type]:
                    try:
                        await callback(data['event'])
                    except Exception as err:
                        _LOGGER.error('Error in event handler: %s', err)
                        
    async def _make_request(self, method: str, endpoint: str, **kwargs) -> Any:
        """Make an HTTP request to the Home Assistant API."""
        url = f"{self._base_url}/api/{endpoint.lstrip('/')}"
        
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers'].update(self._headers)
        
        try:
            async with self._session.request(method, url, **kwargs) as resp:
                resp.raise_for_status()
                if resp.status == 204:  # No content
                    return None
                return await resp.json()
        except aiohttp.ClientError as err:
            _LOGGER.error('Request failed: %s', err)
            raise HomeAssistantError(f'Request failed: {err}') from err
            
    async def get_entities(self) -> List[Dict[str, Any]]:
        """Get all entities from Home Assistant."""
        return await self._make_request('GET', 'states')
        
    async def get_areas(self) -> List[Dict[str, Any]]:
        """Get all areas from Home Assistant."""
        return await self._make_request('GET', 'config/area_registry')
        
    async def get_entity_registry(self) -> List[Dict[str, Any]]:
        """Get the entity registry from Home Assistant."""
        return await self._make_request('GET', 'config/entity_registry')
        
    async def get_services(self) -> Dict[str, Any]:
        """Get all available services from Home Assistant."""
        return await self._make_request('GET', 'services')
        
    async def call_service(
        self,
        domain: str,
        service: str,
        service_data: Optional[Dict[str, Any]] = None,
        target: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Call a service in Home Assistant."""
        data = {}
        if service_data:
            data["service_data"] = service_data
        if target:
            data["target"] = target
            
        return await self._make_request(
            'POST',
            f'services/{domain}/{service}',
            json=data
        )

async def get_hass_info(
    url: str,
    token: str,
    session: Optional[aiohttp.ClientSession] = None,
) -> Things:
    """Get exposed entities and areas from Home Assistant."""
    things = Things()
    
    async with HomeAssistantClient(url, token, session) as client:
        # Get all entities and their states
        entities = await client.get_entities()
        entity_registry = await client.get_entity_registry()
        areas = await client.get_areas()
        services = await client.get_services()
        
        # Process areas
        area_map = {area["area_id"]: area for area in areas}
        
        for area in areas:
            area_names = []
            if "name" in area and area["name"]:
                area_names.append(area["name"])
                
            if area_names:
                things.areas.append(
                    Area(area_id=area["area_id"], names=area_names)
                )
        
        # Process entities
        entity_registry_map = {
            entry["entity_id"]: entry for entry in entity_registry
        }
        
        for entity_state in entities:
            entity_id = entity_state["entity_id"]
            domain = entity_id.split(".", 1)[0]
            
            # Skip entities from ignored domains
            if domain not in {"light", "switch", "media_player", "fan", "cover", "climate"
            }:
                continue
                
            # Get names from entity registry
            names = []
            registry_entry = entity_registry_map.get(entity_id, {})
            
            # Add friendly name if available
            name = registry_entry.get("name") or entity_state.get("attributes", {}).get("friendly_name")
            if name:
                names.append(name)
                
            # Add area name if available
            area_id = registry_entry.get("area_id")
            if area_id and area_id in area_map:
                area = area_map[area_id]
                if "name" in area and area["name"]:
                    names.append(f"{area['name']} {name}" if name else area["name"])
            
            if not names:
                continue
                
            # Create entity with basic info
            entity = Entity(
                entity_id=entity_id,
                names=names,
                domain=domain
            )
            
            # Add domain-specific features
            attributes = entity_state.get("attributes", {})
            
            if domain == "light":
                supported_features = attributes.get("supported_features", 0)
                entity.light_supports_color = bool(supported_features & 16)  # SUPPORT_COLOR
                entity.light_supports_brightness = bool(supported_features & 1)  # SUPPORT_BRIGHTNESS
                
            elif domain == "fan":
                supported_features = attributes.get("supported_features", 0)
                entity.fan_supports_speed = bool(supported_features & FAN_SET_SPEED)
                
            elif domain == "cover":
                supported_features = attributes.get("supported_features", 0)
                entity.cover_supports_position = bool(supported_features & COVER_SET_POSITION)
                
            elif domain == "media_player":
                supported_features = attributes.get("supported_features", 0)
                entity.media_player_supports_pause = bool(supported_features & MEDIA_PLAYER_PAUSE)
                entity.media_player_supports_volume_set = bool(supported_features & MEDIA_PLAYER_VOLUME_SET)
                entity.media_player_supports_next_track = bool(supported_features & MEDIA_PLAYER_NEXT_TRACK)
            
            things.entities.append(entity)
    
    return things
